print("""from zzgui.qt5.zzapp import ZzApp
from zzgui.qt5.zzform import ZzForm as ZzForm
from zzgui.qt5.zzform import zzMess


class firstApp(ZzApp):
    def on_init(self):
        self.add_menu(
            "File|About", lambda: zzMess("First application!"), toolbar=1
        )
        self.add_menu("File|First Form", self.first_form, toolbar=1)
        self.add_menu("File|-")
        self.add_menu("File|Exit", self.close, toolbar=1)
        return super().on_init()

    def first_form(self):
        form = ZzForm("FirstForm")
        form.add_control("", "First Label")
        form.add_control("field", "First Field")
        form.add_control("", "Close Form", control="button", valid=form.close)
        form.show_mdi_modal_form()


if __name__ == "__main__":
    firstApp("zzgui - the first app").run()

""")



import sys

if __name__ == "__main__":

    sys.path.insert(0, ".")

    from demo.demo import demo

    demo()

from PyQt5.QtWidgets import QTabBar, QTabWidget

from zzgui.qt5.zzwindow import ZzFrame
from zzgui.qt5.zzwidget import ZzWidget


class ZzTabBar(QTabBar):
    def get_text(self):
        return self.tabText(self.currentIndex())


class zztab(QTabWidget, ZzWidget, ZzFrame):
    def __init__(self, meta):
        super().__init__(meta)
        self.setTabBar(ZzTabBar())
        self.meta = meta

    def add_tab(self, widget, text=""):
        self.addTab(widget, text)

    def get_text(self):
        return self.tabBar().tabText(self.currentIndex())



import sys

if __name__ == "__main__":

    sys.path.insert(0, ".")

    from demo.demo import demo

    demo()

from PyQt5.QtWidgets import QToolButton, QSizePolicy

from zzgui.qt5.zzwidget import ZzWidget


class zztoolbutton(QToolButton, ZzWidget):
    def __init__(self, meta):
        super().__init__(meta)
        self.setSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum)
        self.set_text(meta.get("label"))
        if self.meta.get("valid"):
            self.clicked.connect(self.valid)



if __name__ == "__main__":
    import sys

    sys.path.insert(0, ".")

    from demo.demo import demo

    demo()


# from zzgui import zzform

import os


from PyQt5.QtWidgets import (
    QApplication,
    QWidget,
    QMainWindow,
    QToolButton,
    QToolBar,
    QFileDialog,
    QTabWidget,
    QTabBar,
    QMdiArea,
    QSizePolicy,
    qApp,
)

from PyQt5.QtCore import QEvent, Qt
from PyQt5.QtGui import QFontMetrics

from zzgui.qt5.zzwindow import ZzQtWindow
from zzgui.qt5.zzwindow import layout
import zzgui.zzapp as zzapp


class ZzApp(zzapp.ZzApp, QApplication):
    def __init__(self, title=""):
        QApplication.__init__(self, [])
        super().__init__(title, main_window_class=ZzMainWindow)
        qApp.focusChanged.connect(self.focus_changed)

    def show_form(self, form=None, modal="modal"):
        if modal == "":  # mdiarea normal window
            self.main_window.zz_tabwidget.currentWidget().addSubWindow(form)
            form.show()
        else:  # mdiarea modal window
            prev_focus_widget = qApp.focusWidget()
            prev_mdi_window = self.main_window.zz_tabwidget.currentWidget().activeSubWindow()
            prev_tabbar_text = self.get_tabbar_text()

            if prev_mdi_window:
                prev_mdi_window.setDisabled(True)

            self.main_window.zz_tabwidget.currentWidget().addSubWindow(form)

            self.set_tabbar_text(form.window_title)

            if modal == "super":  # real modal dialog
                self.disable_toolbar(True)
                self.disable_menubar(True)
                self.disable_tabbar(True)

            form.exec_()
            # form.show()

            if modal == "super":  # real modal dialog
                self.disable_toolbar(False)
                self.disable_menubar(False)
                self.disable_tabbar(False)

            if prev_mdi_window:
                prev_mdi_window.setEnabled(True)

            if prev_focus_widget is not None:
                prev_focus_widget.setFocus()
            self.set_tabbar_text(prev_tabbar_text)

    def build_menu(self):
        self.menu_list = super().reorder_menu(self.menu_list)
        self._main_menu = {}
        QMainWindow.menuBar(self.main_window).clear()
        QMainWindow.menuBar(self.main_window).show()
        for x in self.menu_list:
            _path = x["TEXT"]
            if _path == "" or _path in self._main_menu:
                continue
            prevNode = "|".join(_path.split("|")[:-1])
            topic = _path.split("|")[-1]
            if _path.count("|") == 0:  # first in chain - menu bar
                node = QMainWindow.menuBar(self.main_window)
            else:
                node = self._main_menu[prevNode]
            if _path.endswith("-"):
                node.addSeparator()
            elif x["WORKER"]:
                self._main_menu[_path] = node.addAction(topic)
                self._main_menu[_path].triggered.connect(x["WORKER"])
                if x["TOOLBAR"]:
                    button = QToolButton(self.main_window)
                    button.setText(topic)
                    button.setDefaultAction(self._main_menu[_path])
                    self.main_window.zz_toolbar.addAction(self._main_menu[_path])
            else:
                self._main_menu[_path] = node.addMenu(topic)
        # Show as context menu
        self.main_window.setContextMenuPolicy(Qt.ActionsContextMenu)
        self.main_window.addActions(self.main_window.menuBar().actions())

    def focus_widget(self):
        return qApp.focusWidget()

    def set_style(self):
        if os.path.isfile(self.style_file):
            try:
                with open(self.style_file) as style_data:
                    self.setStyleSheet(style_data.read())
            except Exception:
                print(f"File {self.style_file} reading error...")

    def lock(self):
        self.main_window.menuBar().setDisabled(True)
        self.main_window.zz_toolbar.setDisabled(True)
        self.main_window.zz_tabwidget.setDisabled(True)

    def unlock(self):
        self.main_window.menuBar().setDisabled(False)
        self.main_window.zz_toolbar.setDisabled(False)
        self.main_window.zz_tabwidget.setDisabled(False)

    def process_events(self):
        self.processEvents()

    def show_menubar(self, mode=True):
        zzapp.ZzApp.show_menubar(self)
        if mode:
            QMainWindow.menuBar(self.main_window).show()
        else:
            QMainWindow.menuBar(self.main_window).hide()

    def is_menubar_visible(self):
        return QMainWindow.menuBar(self.main_window).isVisible()

    def show_toolbar(self, mode=True):
        zzapp.ZzApp.show_toolbar(self)
        if mode:
            self.main_window.zz_toolbar.show()
        else:
            self.main_window.zz_toolbar.hide()

    def disable_toolbar(self, mode=True):
        self.main_window.zz_toolbar.setDisabled(True if mode else False)

    def disable_menubar(self, mode=True):
        QMainWindow.menuBar(self.main_window).setDisabled(True if mode else False)

    def disable_tabbar(self, mode=True):
        self.main_window.zz_tabwidget.tabBar().setDisabled(True if mode else False)

    def is_toolbar_visible(self):
        return self.main_window.zz_toolbar.isVisible()

    def show_tabbar(self, mode=True):
        zzapp.ZzApp.show_tabbar(self)
        if mode:
            self.main_window.zz_tabwidget.tabBar().show()
        else:
            self.main_window.zz_tabwidget.tabBar().hide()

    def is_tabbar_visible(self):
        return self.main_window.zz_tabwidget.tabBar().isVisible()

    def get_tabbar_text(self):
        return self.main_window.zz_tabwidget.tabBar().tabText(self.main_window.zz_tabwidget.currentIndex())

    def show_statusbar_mess(self, text=""):
        self.main_window.statusBar().showMessage(f"{text}")

    def set_tabbar_text(self, text=""):
        self.main_window.zz_tabwidget.tabBar().setTabText(self.main_window.zz_tabwidget.currentIndex(), text)

    def show_statusbar(self, mode=True):
        zzapp.ZzApp.show_statusbar(self)
        if mode:
            self.main_window.statusBar().show()
        else:
            self.main_window.statusBar().hide()

    def is_statusbar_visible(self):
        return self.main_window.statusBar().isVisible()

    def get_char_width(self, char="w"):
        return QFontMetrics(self.font()).width(char)

    def get_char_height(self):
        return QFontMetrics(self.font()).height()

    @staticmethod
    def get_open_file_dialoq(header="Open file", path="", filter=""):
        return QFileDialog.getOpenFileName(None, header, path, filter)

    @staticmethod
    def get_save_file_dialoq(header="Save file", path="", filter=""):
        return QFileDialog.getSaveFileName(None, header, path, filter)

    def run(self):
        self.main_window.restore_geometry(self.settings)
        self.main_window.show()
        super().run()
        self.exec_()


class ZzMainWindow(QMainWindow, zzapp.ZzMainWindow, ZzQtWindow):
    class ZzTabWidget(QTabWidget):
        def __init__(self, parent):
            super().__init__(parent)
            self.addTab(QWidget(parent), "")
            self.setAttribute(Qt.WA_DeleteOnClose)
            self.addTabButton = QToolButton(self)
            self.addTabButton.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Minimum)
            self.addTabButton.setText("+")
            self.addTabButton.clicked.connect(self.addTab)
            self.tabBar().setTabButton(0, QTabBar.RightSide, self.addTabButton)
            self.tabBar().setTabEnabled(0, False)

            self.closeButton = QToolButton(self)
            self.closeButton.setText("x")
            self.closeButton.clicked.connect(self.closeSubWindow)
            self.setCornerWidget(self.closeButton)
            self.currentChanged.connect(self._currentChanged)

            self.addTab()
            self.setCurrentIndex(0)

        def _currentChanged(self, index: int):
            # bug path when subwindow in tab 0 lost focus if we close subwindow in other tab
            if index == 0 and self.currentWidget().subWindowList():
                self.currentWidget().subWindowList()[-1].setFocus()

        def closeSubWindow(self):
            currentTabIndex = self.currentIndex()
            if self.currentWidget().activeSubWindow():
                self.currentWidget().activeSubWindow().close()
            elif self.count() > 2:  # close tab if them >2
                self.setCurrentIndex(currentTabIndex - 1)
                self.removeTab(currentTabIndex)

        def addTab(self, widget=None, label="="):
            if not widget:
                widget = QMdiArea(self)
                widget.setOption(QMdiArea.DontMaximizeSubWindowOnActivation)
                widget.setHorizontalScrollBarPolicy(Qt.ScrollBarAsNeeded)
                widget.setVerticalScrollBarPolicy(Qt.ScrollBarAsNeeded)

            self.insertTab(self.count() - 1, widget, label)
            self.setCurrentIndex(self.count() - 2)

    def __init__(self, title=""):
        self._core_app = QApplication([])
        super().__init__()
        self.zz_toolbar = QToolBar(self)
        self.zz_tabwidget = self.ZzTabWidget(self)
        self.setCentralWidget(QWidget(self))
        self.centralWidget().setLayout(layout("v"))
        self.centralWidget().layout().addWidget(self.zz_toolbar)
        self.centralWidget().layout().addWidget(self.zz_tabwidget)
        self.statusBar().setVisible(True)
        self.set_title(title)

    def show(self):
        QMainWindow.show(self)

    def closeEvent(self, event: QEvent):
        event.accept()
        zzapp.zz_app.close()

    def close(self):
        QMainWindow.close(self)
        super().close()
        os._exit(0)



if __name__ == "__main__":
    import sys

    sys.path.insert(0, ".")

    from demo.demo import demo

    demo()

from PyQt5.QtWidgets import QWidget, QHBoxLayout, QVBoxLayout, QFormLayout, QGridLayout

from PyQt5.QtCore import Qt

from zzgui import zzwindow

zz_align = {
    "": Qt.AlignLeft | Qt.AlignTop,
    "-1": Qt.AlignLeft | Qt.AlignTop,
    "0": Qt.AlignLeft | Qt.AlignTop,
    "1": Qt.AlignLeft | Qt.AlignBottom,
    "2": Qt.AlignHCenter | Qt.AlignBottom,
    "3": Qt.AlignRight | Qt.AlignBottom,
    "4": Qt.AlignLeft | Qt.AlignVCenter,
    "5": Qt.AlignHCenter | Qt.AlignVCenter,
    "6": Qt.AlignRight | Qt.AlignVCenter,
    "7": Qt.AlignLeft | Qt.AlignTop,
    "8": Qt.AlignHCenter | Qt.AlignTop,
    "9": Qt.AlignRight | Qt.AlignTop,
}


def layout(arg="h"):
    if arg.lower().startswith("v"):
        layout = QVBoxLayout()
        layout.setAlignment(zz_align["7"])
    elif arg.lower().startswith("f"):
        layout = QFormLayout()
        layout.setLabelAlignment(zz_align["6"])
        layout.layout().setSpacing(2)
    elif arg.lower().startswith("g"):
        layout = QGridLayout()
    else:
        layout = QHBoxLayout()
        layout.setAlignment(zz_align["7"])
    layout.layout().setContentsMargins(0, 0, 0, 0)
    return layout


class ZzFrame(zzwindow.ZzFrame, QWidget):
    def set_mode(self, mode="v"):
        self.splitter = None
        super().set_mode(mode=mode)
        if self.layout() is not None:
            return
        self.setLayout(layout(mode))

    def insert_widget(self, pos=None, widget=None):
        # if widget:
        #     # widget.setContentsMargins(0,20, 0, 0)
        #     if widget.label:
        #         print("--",widget, widget.label.get_text())
        self.layout().addWidget(widget)

    def add_row(self, label=None, widget=None):
        # if widget:
        #     widget.setContentsMargins(0, 0, 0, 0)
        #     print("f", widget, widget.label.get_text())
        self.layout().addRow(label, widget)


class ZzQtWindow(zzwindow.ZzWindow, ZzFrame):
    def __init__(self, title=""):
        super().__init__()
        self.set_title(title)

    def set_position(self, left, top):
        self.move(left, top)
        # self.move(100, 100)

    def set_size(self, width, height):
        self.resize(width, height)

    def get_position(self):
        return (self.pos().x(), self.pos().y())

    def get_size(self):
        if hasattr(self, "parent") and self.parent() is not None:
            return (self.parent().size().width(), self.parent().size().height())
        else:
            return (self.size().width(), self.size().height())

    def set_disabled(self, arg=True):
        self.setEnabled(True if not arg else False)

    def set_enabled(self, arg=True):
        self.setEnabled(True if arg else False)

    def set_title(self, title):
        super().set_title(title)
        QWidget.setWindowTitle(self, title)

    def is_maximized(self):
        return 1 if QWidget.isMaximized(self) else 0

    def show_maximized(self):
        QWidget.showMaximized(self)



import sys

if __name__ == "__main__":

    sys.path.insert(0, ".")

    from demo.demo import demo

    demo()


from PyQt5.QtWidgets import (
    QTableView,
    QStyledItemDelegate,
    QAbstractItemView,
    QStyle,
    QStyleOptionButton,
    QApplication,
)
from PyQt5.QtGui import QPalette, QPainter

from PyQt5.QtCore import Qt, QAbstractTableModel, QVariant

from zzgui.qt5.zzwindow import zz_align
from zzgui.zzutils import int_
from zzgui.zzmodel import ZzModel
from zzgui.qt5.widgets.zzlookup import zzlookup


class zzDelegate(QStyledItemDelegate):
    def paint(self, painter: QPainter, option, index):
        if self.parent().currentIndex().column() == index.column():
            color = option.palette.color(QPalette.AlternateBase).darker(900)
            color.setAlpha(color.alpha() / 10)
            painter.fillRect(option.rect, color)
        meta = self.parent().model().zz_model.meta[index.column()]
        if meta.get("control") == "check":
            self.paint_checkbox(painter, option, index, meta)
            return
        # elif meta.get("relation"):
        #     super().paint(painter, option, index)
        #     self.paint_relation_button(painter, option, index, meta)
        #     return
        super().paint(painter, option, index)

    # def paint_relation_button(self, painter: QPainter, option, index, meta):
    #     pb_option = QStyleOptionButton()
    #     pb_option.text = "?"
    #     checkBoxRect = QApplication.style().subElementRect(QStyle.SE_PushButtonBevel, pb_option, None)
    #     sz = 30
    #     pb_option.rect = option.rect
    #     pb_option.rect.setX(pb_option.rect.x() - sz + option.rect.width())
    #     print(checkBoxRect.height())
    #     pb_option.rect.setHeight(sz)
    #     pb_option.rect.setWidth(sz)
    #     QApplication.style().drawControl(QStyle.CE_PushButton, pb_option, painter)

    def paint_checkbox(self, painter: QPainter, option, index, meta):
        """paint checkbox - left - with top+left alignment"""
        if meta.get("num"):
            checked = True if int_(index.data()) else False
        else:
            checked = True if index.data() else False
        cb_option = QStyleOptionButton()
        if checked:
            cb_option.state |= QStyle.State_On
        else:
            cb_option.state |= QStyle.State_Off
        checkBoxRect = QApplication.style().subElementRect(QStyle.SE_CheckBoxIndicator, cb_option, None)
        cb_option.rect = option.rect
        cb_option.rect.setX(cb_option.rect.x() + checkBoxRect.width() / 2)
        if cb_option.rect.height() > checkBoxRect.height() * 2 + 3:
            cb_option.rect.setHeight(checkBoxRect.height() * 2)
        QApplication.style().drawControl(QStyle.CE_CheckBox, cb_option, painter)


class zzgrid(QTableView):
    class ZzTableModel(QAbstractTableModel):
        def __init__(self, zz_model):
            super().__init__(parent=None)
            self.zz_model: ZzModel = zz_model
            self._zz_model_refresh = self.zz_model.refresh
            self.zz_model.refresh = self.refresh

        def set_order(self, column):
            self.zz_model.order_column(column)

        def rowCount(self, parent=None):
            return self.zz_model.row_count()

        def columnCount(self, parent=None):
            return self.zz_model.column_count()

        def refresh(self):
            self.beginResetModel()
            self.endResetModel()
            self._zz_model_refresh()

        def data(self, index, role=Qt.DisplayRole):
            if role == Qt.DisplayRole:
                return QVariant(self.zz_model.data(index.row(), index.column()))
            elif role == Qt.TextAlignmentRole:
                return QVariant(zz_align[str(self.zz_model.alignment(index.column()))])
            else:
                return QVariant()

        def headerData(self, col, orientation, role=Qt.DisplayRole):
            if orientation == Qt.Horizontal and role == Qt.DisplayRole:
                return self.zz_model.headers[col]
            elif orientation == Qt.Vertical and role == Qt.DisplayRole:
                return QVariant("")
            else:
                return QVariant()

    # currentCellChangedSignal = pyqtSignal(int, int)

    def __init__(self, meta):
        super().__init__()
        self.meta = meta

        self.zz_form = self.meta.get("form")
        self.zz_model = self.zz_form.model

        # self.setModel(self.ZzTableModel(self.zz_form.model))
        self.setItemDelegate(zzDelegate(self))
        self.setTabKeyNavigation(False)

        self.horizontalHeader().setSectionsMovable(True)
        self.setAlternatingRowColors(True)
        self.setSelectionBehavior(QAbstractItemView.SelectRows)
        self.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.setContextMenuPolicy(Qt.ActionsContextMenu)
        self.horizontalHeader().setDefaultAlignment(zz_align["7"])
        self.doubleClicked.connect(self.zz_form.grid_double_clicked)
        self.horizontalHeader().sectionClicked.connect(self.zz_form.grid_header_clicked)
        self.setModel(self.ZzTableModel(self.zz_form.model))

    def currentChanged(self, current, previous):
        # self.currentCellChangedSignal.emit(current.row(), current.column())
        super().currentChanged(current, previous)
        self.model().dataChanged.emit(current, previous)
        self.zz_form.grid_index_changed(self.currentIndex().row(), self.currentIndex().column())

    def current_index(self):
        return self.currentIndex().row(), self.currentIndex().column()

    def set_focus(self):
        self.setFocus()

    def row_count(self):
        return self.model().rowCount()

    def column_count(self):
        return self.model().columnCount()

    def set_index(self, row, column=None):
        if row < 0:
            row = 0
        elif row > self.row_count() - 1:
            row = self.row_count() - 1

        if column is None:
            column = self.currentIndex().column()
        elif column < 0:
            column = 0
        elif column > self.column_count() - 1:
            column = self.column_count() - 1

        self.setCurrentIndex(self.model().index(row, column))

    def keyPressEvent(self, event):
        event.accept()
        # if ev.key() in [Qt.Key_F] and ev.modifiers() == Qt.ControlModifier:
        #     self.searchText()
        # if event.key() in [Qt.Key_Asterisk]:
        if (
            event.text()
            and event.key() not in (Qt.Key_Escape, Qt.Key_Enter, Qt.Key_Return, Qt.Key_Space)
            and self.model().rowCount() >= 1
            and event.modifiers() != Qt.ControlModifier
            and event.modifiers() != Qt.AltModifier
        ):
            lookup_widget = zz_grid_lookup(self, event.text())
            lookup_widget.show(self, self.currentIndex().column())
        else:
            super().keyPressEvent(event)

    def get_selected_rows(self):
        return [x.row() for x in self.selectionModel().selectedRows()]

    def get_columns_headers(self):
        rez = {}
        hohe = self.horizontalHeader()
        for x in range(0, hohe.count()):
            rez[hohe.model().headerData(x, Qt.Horizontal, Qt.DisplayRole)] = x
        return rez

    def get_columns_settings(self):
        rez = []
        hohe = self.horizontalHeader()
        for x in range(0, hohe.count()):
            header = hohe.model().headerData(x, Qt.Horizontal, Qt.DisplayRole)
            width = self.columnWidth(x)
            pos = hohe.visualIndex(x)
            rez.append({"name": header, "data": f"{pos}, {width}"})
        return rez

    def set_column_settings(self, col_settings):
        headers = self.get_columns_headers()
        for x in col_settings:
            if "," not in col_settings[x]:
                continue
            column_pos = int_(col_settings[x].split(",")[0])
            column_width = int_(col_settings[x].split(",")[1])
            # column_width = column_width if column_width else 10
            self.setColumnWidth(headers.get(x), column_width)
            old_visual = self.horizontalHeader().visualIndex(int_(headers[x]))
            self.horizontalHeader().moveSection(old_visual, column_pos)
        self.set_index(0, self.horizontalHeader().logicalIndex(0))


class zz_grid_lookup(zzlookup):
    def lookup_list_selected(self):
        self.zz_grid.set_index(self.found_rows[self.lookup_list.currentRow()][0])
        self.close()

    def lookup_search(self):
        self.lookup_list.clear()
        self.found_rows = self.zz_model.lookup(self.zz_model_column, self.lookup_edit.get_text())
        for x in self.found_rows:
            self.lookup_list.addItem(f"{x[1]}")

    def show(self, zz_grid, column):
        self.zz_grid = zz_grid
        self.zz_model_column = column
        self.zz_model = zz_grid.zz_model
        return super().show()

    def set_geometry(self):
        parent = self.parent()
        rect = parent.visualRect(parent.currentIndex())
        rect.moveTop(parent.horizontalHeader().height() + 2)
        rect.moveLeft(parent.verticalHeader().width() + rect.x() + 2)
        pos = rect.topLeft()
        pos = parent.mapToGlobal(pos)
        self.setFixedWidth(parent.width() - rect.x())
        self.move(pos)



import sys

if __name__ == "__main__":

    sys.path.insert(0, ".")

    from demo.demo import demo

    demo()

from PyQt5.QtWidgets import QCheckBox, QSizePolicy

from zzgui.qt5.zzwidget import ZzWidget
from zzgui.zzutils import int_

class zzcheck(QCheckBox, ZzWidget):
    def __init__(self, meta):
        super().__init__(meta)
        self.setText(meta.get("pic", meta.get("label", "")))
        self.managed_widgets = []
        self.setSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum)
        self.stateChanged.connect(self.state_changed)

    def state_changed(self):
        for x in self.managed_widgets:
            x.set_enabled(self.isChecked())
            if x.is_enabled():
                x.set_focus()
        self.valid()

    def add_managed_widget(self, widget):
        self.managed_widgets.append(widget)
        widget.check = self

    def remove_managed_widget(self, widget):
        if widget in self.managed_widgets:
            self.managed_widgets.pop(self.managed_widgets.index(widget))

    def set_text(self, text):
        if self.meta.get("num"):
            self.setChecked(True if int_(text) else False)
        else:
            self.setChecked(True if text else False)

    def get_text(self):
        if self.meta.get("num"):
            return 1 if self.isChecked() else 0
        else:
            return "*" if self.isChecked() else ""

    def set_checked(self, mode=True):
        self.set_text(mode)

    def is_checked(self):
        return True if self.get_text() else False



if __name__ == "__main__":
    import sys

    sys.path.insert(0, ".")

    from demo.demo import demo

    demo()

import csv
import datetime
import re

import zzgui.zzapp as zzapp
from zzgui.zzutils import num
from zzdb.cursor import ZzCursor
from zzdb.db import ZzDb


class ZzModel:
    def __init__(self):
        self.zz_form = None
        self.columns = []
        self.headers = []
        self.alignments = []

        self.records = []
        self.proxy_records = []
        self.use_proxy = False
        self.relation_cache = {}
        self.lastdata_error_text = ""

        self.meta = []

        # CRUD flags
        self.readonly = True

        self.delete_enabled = False
        self.insert_enabled = False
        self.update_enabled = False

        self.filterable = False
        
        self.data_changed = False

        self.order_text = ""
        self.where_text = ""

    def get_table_name(self):
        return ""

    def get_data_error(self):
        return self.lastdata_error_text

    def set_data_error(self, text=""):
        self.lastdata_error_text = text

    def insert(self, record: dict, current_row=0):
        # print(record)
        return True

    def update(self, record: dict, current_row=0):
        # print(record)
        return True

    def delete(self, current_row=0):
        # print(current_row)
        return True

    def set_where(self, where_text=""):
        self.where_text = where_text

    def get_where(self):
        return self.where_text

    def set_order(self, order_data=""):
        if isinstance(order_data, int):
            self.order_text = self.columns[order_data]
        elif isinstance(order_data, list):
            self.order_text = ",".join(order_data)
        else:
            self.order_text = order_data

    def refresh(self):
        self.relation_cache = {}

    def reset(self):
        self.records = []
        self.proxy_records = []
        self.use_proxy = False
        self.relation_cache = {}
        self.lastdata_error_text = ""

    def set_records(self, records):
        self.records = records

    def build(self):
        self.columns = []
        self.headers = []
        self.alignments = []
        self.meta = []
        for meta in self.zz_form.controls:
            if meta.get("name", "").startswith("/") or meta.get("nogrid"):
                continue
            if meta.get("control", "") in ["button", "widget", "form"]:
                continue
            self.zz_form.model.add_column(meta)

    def add_column(self, meta):
        if not meta.get("control"):
            meta["control"] = "line"

        meta = zzapp.ZzControls.validate(meta)
        self.columns.append(meta["name"])
        self.headers.append(meta["label" if meta.get("saygrid") else "label"])
        self.alignments.append(meta.get("alignment", "7"))
        self.meta.append(meta)

    def get_record(self, row):
        if self.use_proxy:
            return self.records[self.proxy_records[row]]
        else:
            return self.records[row]

    def _get_related(self, value, meta, do_not_show_value=False, reset_cache=False):
        if meta.get("num") and num(value) == 0:
            return ""
        elif value == "":
            return ""
        key = (meta["to_table"], f"{meta['to_column']}='{value}'", meta["related"])
        if not reset_cache and key in self.relation_cache:
            related = self.relation_cache[key]
        else:
            related = self.get_related(meta["to_table"], f"{meta['to_column']}='{value}'", meta["related"])
            self.relation_cache[key] = related
        if related is None:
            related = ""
        if do_not_show_value:
            return f"{related}"
        else:
            return f"{value},{related}"

    def get_related(self, to_table, filter, related):
        return "get_related"

    def data(self, row, col, role="display"):
        if role == "display":
            colName = self.columns[col]
            value = self.get_record(row).get(colName, "")
            meta = self.meta[col]
            if meta.get("relation"):
                value = self._get_related(value, meta)
            elif self.is_strign_for_num(meta):
                value = meta.get("pic").split(";")[int(num(value)) - 1]
            elif meta.get("num"):  # Numeric value
                if num(value) == 0:  # do not show zero
                    value = ""
                elif meta.get("pic") == "F":  # financial format
                    format_string = zzapp.FINANCIAL_FORMAT % meta.get("datadec", 0)
                    value = format_string.format(num(value))
            elif meta["datatype"] == "date":
                try:
                    value = datetime.datetime.strptime(value, "%Y-%m-%d").strftime(zzapp.DATA_FORMAT_STRING)
                except Exception:
                    value = ""
            return value

    def is_strign_for_num(self, meta):
        """return str data from numeric controls - radio, list, combo"""
        return meta.get("num") and ("radio" in meta["control"] or meta["control"] in ("list", "combo"))

    def alignment(self, col):
        if self.meta[col].get("relation"):
            return 7
        elif self.is_strign_for_num(self.meta[col]):
            return 7
        return self.alignments[col]

    def column_header_data(self, col):
        return self.headers[col]

    def row_header_data(self, row):
        return f"{row}"

    def row_count(self):
        if self.use_proxy:
            return len(self.proxy_records)
        else:
            return len(self.records)

    def column_count(self):
        return len(self.columns)

    def parse_lookup_text(self, text):
        text = text.upper()
        raw_cond_list = ["+"] + re.split("([+-])", text)
        cond_list = []
        for cond in range(int(len(raw_cond_list) / 2)):
            operator = raw_cond_list[cond * 2]
            value = raw_cond_list[cond * 2 + 1]
            if operator and value:
                cond_list.append([operator, value])
        return cond_list

    def lookup(self, column, text):
        """search for text in column, return list of [row, value]"""
        cond_list = self.parse_lookup_text(text)
        rez = []
        for x in range(self.row_count()):
            cond_result = True
            for cond in cond_list:
                operator = cond[0]
                value = cond[1]
                model_value = self.get_record(x)[self.columns[column]]
                if self.meta[column].get("relation"):
                    model_value = self._get_related(model_value, self.meta[column])

                if operator == "+" and value not in model_value.upper():
                    cond_result = False
                elif operator == "-" and value in model_value.upper():
                    cond_result = False
            if cond_result:
                rez.append([x, model_value])
        return rez

    def data_export(self, file):
        pass

    def data_import(self, file):
        pass


class ZzCsvModel(ZzModel):
    def __init__(self, csv_file_object=None):
        super().__init__()
        csv_dict = csv.DictReader(csv_file_object)
        # If there are names with space -  replace spaces in columns names
        if [filename for filename in csv_dict.fieldnames if " " in filename]:
            fieldnames = [x.replace(" ", "_") for x in csv_dict.fieldnames]
            csv_dict = csv.DictReader(csv_file_object, fieldnames)
        self.set_records([x for x in csv_dict])
        self.filterable = True

    def update(self, record: dict, current_row):
        self.records[current_row] = record
        self.data_changed = True
        self.refresh()
        return True

    def insert(self, record: dict, current_row):
        self.records.append(record)
        self.data_changed = True
        self.refresh()
        return True

    def delete(self, row_number):
        self.records.pop(row_number)
        self.data_changed = True
        self.refresh()
        return True

    def set_where(self, where_text=""):
        self.where_text = where_text
        if self.where_text:
            self.use_proxy = True
            self.proxy_records = []
            for row, rec in enumerate(self.records):
                if eval(self.where_text, rec):
                    self.proxy_records.append(row)
        else:
            self.use_proxy = False
        self.refresh()

    def set_order(self, order_data=""):
        super().set_order(order_data=order_data)

        colname = self.order_text

        if self.proxy_records:
            sort_records = {}
            for x in range(len(self.proxy_records)):
                value = self.records[self.proxy_records[x]][colname]
                if value not in sort_records:
                    sort_records[value] = [self.proxy_records[x]]
                else:
                    sort_records[value].append(self.proxy_records[x])
        else:
            sort_records = {}
            for x in range(len(self.records)):
                if self.records[x][colname] not in sort_records:
                    sort_records[self.records[x][colname]] = [x]
                else:
                    sort_records[self.records[x][colname]].append(x)

        sorted_keys = sorted([x for x in sort_records.keys()])
        # sorted_keys.sort()
        tmp_proxy_records = []
        for x in sorted_keys:
            for y in sort_records[x]:
                tmp_proxy_records.append(y)

        self.proxy_records = tmp_proxy_records
        self.use_proxy = True


class ZzCursorModel(ZzModel):
    def __init__(self, cursor: ZzCursor = None):
        super().__init__()
        self.cursor = cursor

        self.readonly = False
        self.delete_enabled = False
        self.insert_enabled = False
        self.update_enabled = False

    def get_table_name(self):
        return self.cursor.table_name

    def row_count(self):
        return self.cursor.row_count()

    def get_record(self, row):
        return self.cursor.record(row)

    def refresh(self):
        super().refresh()
        self.cursor.refresh()

    def delete(self, current_row=0):
        self.set_data_error()
        record = self.get_record(current_row)
        if self.cursor.delete(record, refresh=False):
            return True
        else:
            self.set_data_error(self.cursor.last_sql_error())
            return False

    def update(self, record: dict, current_row=0):
        self.set_data_error()
        if self.cursor.update(record):
            self.refresh()
            return True
        else:
            self.set_data_error(self.cursor.last_sql_error())
            return False

    def insert(self, record: dict, current_row=0):
        self.set_data_error()
        if self.cursor.insert(record):
            self.refresh()
            return True
        else:
            self.set_data_error(self.cursor.last_sql_error())
            return False

    def get_related(self, to_table, filter, related):
        db: ZzDb = self.cursor.zz_db
        return db.get(to_table, filter, related)

    def set_order(self, order_data):
        if self.cursor.table_name:
            super().set_order(order_data=order_data)
            self.cursor.set_order(self.order_text)

    def add_column(self, meta):
        """update metadata from db"""
        db: ZzDb = self.cursor.zz_db
        db_meta = db.db_schema.get_schema_table_attr(self.cursor.table_name, meta["name"])
        meta["pk"] = db_meta.get("pk", "")
        meta["datatype"] = db_meta.get("datatype", meta["datatype"])
        if num(meta["datalen"]) < num(db_meta.get("datalen", 10)):
            meta["datalen"] = int(num(db_meta.get("datalen", 10)))
        if "datadec" not in meta:
            meta["datadec"] = int(num(db_meta.get("datadec", 2)))
        return super().add_column(meta)

    def set_where(self, where_text=""):
        self.cursor.set_where(where_text)
        self.refresh()
        return super().set_where(where_text)

    def data_export(self, file: str):
        if file.lower().endswith(".csv"):
            self.cursor.export_csv(file)
        else:
            self.cursor.export_json(file)

    def data_import(self, file: str):
        if file.lower().endswith(".csv"):
            self.cursor.import_csv(file)
        else:
            self.cursor.import_json(file)
        self.refresh()



import sys

if __name__ == "__main__":

    sys.path.insert(0, ".")

    from demo.demo import demo

    demo()

from PyQt5.QtWidgets import QFrame

from zzgui.zzform import ZzForm
from zzgui.zzutils import num
from zzgui.qt5.zzwidget import ZzWidget
from zzgui.qt5.zzwindow import ZzFrame
from zzgui.qt5.widgets.zzline import zzline
from zzgui.qt5.widgets.zzbutton import zzbutton
from zzgui.qt5.widgets.zzlookup import zzlookup
from zzdb.db import ZzDb


class zzrelation(QFrame, ZzWidget, ZzFrame):
    def __init__(self, meta):
        super().__init__(meta)
        ZzFrame.__init__(self, "h")
        self.meta = meta
        meta["valid"] = self.get_valid

        self.get = zzline(meta)
        self.get.textChanged.connect(self.get_text_changed)
        self.button = zzbutton({"label": "?", "datalen": 3, "valid": self.show_related_form})
        self.say = zzline({"disabled": "*"})
        self.to_form = None
        if self.meta.get("to_form"):
            self.to_form: ZzForm = self.meta.get("to_form")()
            self.to_form.max_child_level = 0
            self.to_form.title += " ."

        self.add_widget(self.get)
        self.add_widget(self.button)
        self.add_widget(self.say)
        self.set_text(self.meta.get("data", ""))
        self.get_valid()

    def show_related_form(self):
        if isinstance(self.to_form, ZzForm):
            self.to_form.add_action("Select", self.show_related_form_result, hotkey="Enter", tag="select")

            def seek():
                row = self.to_form.model.cursor.seek_row({self.meta["to_column"]: self.get_text()})
                self.to_form.set_grid_index(row)

            self.to_form._after_grid_create = seek
            self.to_form.show_mdi_modal_grid()
            # self.set_related()

    def show_related_form_result(self):
        if self.to_form:
            self.get.set_text(self.to_form.r.__getattr__(self.meta["to_column"]))
            self.to_form.close()
            self.get.set_focus()
            self.get_valid()

    def get_valid(self):
        value = self.get.get_text()
        if self.meta.get("num") and num(value) == 0:
            return True
        elif value == "":
            return True
        return self.set_related()

    def set_related(self):
        if self.meta["form"].model:
            rel = self.meta["form"].model._get_related(
                self.get.text(), self.meta, do_not_show_value=1, reset_cache=1
            )
        elif self.meta["form"].db:  # datasource provided
            rel = self.meta["form"].db.get(
                self.meta["to_table"],
                f"{self.meta['to_column']}='{self.get.text()}'",
                self.meta["related"],
            )
            if rel == {}:
                rel = None
                self.say.set_text("")
                self.show_related_form()
                return True

        if rel is None:
            self.say.set_text("")
            return False
        else:
            self.say.set_text(rel)
            return True

    def set_text(self, text):
        if hasattr(self, "get"):
            self.get.set_text(text)
            self.get_valid()

    def get_text(self):
        return self.get.get_text()

    def get_text_changed(self, text):
        if "*" in text and self.meta.get("num") or text.startswith("*"):
            lookup_widget = zz_realtion_lookup(self, "")
            lookup_widget.show(self.meta)

    def set_focus(self):
        self.get.setFocus()


class zz_realtion_lookup(zzlookup):
    def __init__(self, parent, text):
        super().__init__(parent, text)

    def lookup_search(self):
        self.lookup_list.clear()
        zz_db: ZzDb = self.parent().to_form.model.cursor.zz_db

        sql = "select {to_column} as tocol, {related} as recol from {to_table}".format(**self.meta)

        related = self.meta.get("related")
        cond_list = self.parent().to_form.model.parse_lookup_text(self.lookup_edit.get_text())

        where = " and ".join(
            [f"{related} {'' if x[0] == '+' else 'not'} like '%{x[1]}%' " for x in cond_list]
        )
        sql += f" where {where}"
        cursor = zz_db.cursor(sql)
        self.found_rows = [x for x in cursor.records()]
        for x in self.found_rows:
            self.lookup_list.addItem(f"""{x["recol"]} ({x["tocol"]})""")

    def lookup_list_selected(self):
        self.parent().set_text(self.found_rows[self.lookup_list.currentRow()]["tocol"])
        self.close()

    def set_geometry(self):
        parent = self.parent()
        self.setFixedWidth(parent.width())
        self.move(parent.mapToGlobal(parent.rect().topLeft()))

    def show(self, meta):
        self.meta = meta
        return super().show()



if __name__ == "__main__":
    import sys

    sys.path.insert(0, ".")

    from demo.demo import demo

    demo()

from zzgui.zzutils import num


class ZzWidget:
    def __init__(self, meta={}):
        self.meta = meta
        self.form = None
        self.label = None
        self.check = None
        if self.meta.get("readonly"):
            self.set_readonly(True)
        if self.meta.get("disabled"):
            self.set_disabled(True)
        if self.meta.get("mess"):
            self.set_tooltip(self.meta.get("mess"))
        if hasattr(self, "set_text") and self.meta.get("data"):
            self.set_text(self.meta.get("data"))

        max_width = max(num(self.meta.get("datalen", 0)), len(self.meta.get("pic", "")))
        if max_width:
            self.set_maximum_width(max_width)
        if max_width:
            self.set_maximum_len(max_width)

        self.set_alignment(self.meta.get("alignment", 7))

    def set_readonly(self, arg):
        pass

    def set_disabled(self, arg=True):
        self.set_enabled(not arg)

    def set_enabled(self, arg=True):
        pass

    def set_visible(self, arg=True):
        pass

    def set_tooltip(self, mess):
        pass

    def set_focus(self):
        pass

    def is_enabled(self):
        pass

    def set_text(self, text):
        pass

    def get_text(self):
        pass

    def valid(self):
        if self.meta.get("valid") is not None:
            return self.meta.get("valid")()
        else:
            return True

    def when(self):
        return self.meta.get("when", lambda: True)()

    def set_maximum_len(self, length):
        pass

    def set_maximum_width(self, width):
        pass

    def set_alignment(self, alignment):
        pass



import sys

if __name__ == "__main__":

    sys.path.insert(0, ".")

    from demo.demo import demo

    demo()

from PyQt5.QtWidgets import QSpinBox

from zzgui.qt5.zzwidget import ZzWidget


class zzspin(QSpinBox, ZzWidget):
    def __init__(self, meta):
        super().__init__(meta)
        self.meta = meta
        self.set_text(meta.get("data"))



import sys

if __name__ == "__main__":

    sys.path.insert(0, ".")

    from demo.demo import demo

    demo()


from PyQt5.QtWidgets import QPushButton, QSizePolicy
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QFontMetrics

from zzgui.qt5.zzwidget import ZzWidget
from zzgui.zzutils import num

class zzbutton(QPushButton, ZzWidget):
    def __init__(self, meta):
        super().__init__(meta)
        # self.meta = meta
        self.setSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum)
        self.set_text(meta.get("label"))
        if self.meta.get("valid"):
            self.clicked.connect(self.valid)
        ml = num(self.meta.get("datalen"))
        if ml:
            self.setMinimumWidth(QFontMetrics(self.font()).width("W") * ml)

    def keyPressEvent(self, ev):
        if ev.key() in [Qt.Key_Enter, Qt.Key_Return] and not self.meta.get("eat_enter"):
            ev.accept()
            self.focusNextChild()
        else:
            super().keyPressEvent(ev)



if __name__ == "__main__":
    import sys

    sys.path.insert(0, ".")

    from demo.demo import demo

    demo()


from zzgui import zzmodel

class ZzModel(zzmodel.ZzModel):
    def refresh(self):
        self.beginResetModel()
        self.endResetModel()
        return super().refresh()



if __name__ == "__main__":
    import sys

    sys.path.insert(0, ".")

    from demo.demo import demo

    demo()

from PyQt5.QtWidgets import QDialog, QMdiSubWindow, QApplication
from PyQt5.QtCore import Qt, QEvent
from PyQt5.QtGui import QKeySequence, QKeyEvent


import zzgui.zzapp as zzapp
import zzgui.zzform as zzform
import zzgui.qt5.widgets

from zzgui.qt5.zzapp import ZzQtWindow
from zzgui.zzutils import num

import zzgui.zzdialogs
from zzgui.zzdialogs import zzMess, zzWait, zzAskYN


class ZzForm(zzform.ZzForm):
    def __init__(self, title=""):
        super().__init__(title=title)
        self._ZzFormWindow_class = ZzFormWindow
        self._zzdialogs = zzgui.zzdialogs


class ZzFormWindow(QDialog, zzform.ZzFormWindow, ZzQtWindow):
    def __init__(self, zz_form: ZzForm, title=""):
        super().__init__(zz_form)
        title = title if title else zz_form.title
        ZzQtWindow.__init__(self, title)
        self._widgets_package = zzgui.qt5.widgets

    def restore_geometry(self, settings):
        paw = self.parent()
        if paw is not None:
            # save default == minimal size
            sizeBefore = paw.size()

            width = num(settings.get(self.window_title, "width", "-1"))
            height = num(settings.get(self.window_title, "height", "-1"))
            if -1 in [width, height]:  # bad settings or first run
                width = paw.parent().size().width()*0.9 if self.mode == "grid" else 0.5
                height = paw.parent().size().height()*0.9 if self.mode == "grid" else 0.5
            paw.resize(width, height)

            sizeAfter = paw.size()
            self.expand_size(paw, sizeBefore, sizeAfter)

            left = num(settings.get(self.window_title, "left", "-1"))
            top = num(settings.get(self.window_title, "top", "-1"))
            if -1 in [left, top]:  # bad settings or first run
                left, top = self.center_pos()

            paw.move(left, top)

            self.fit_size_and_pos(paw)

            if num(settings.get(self.window_title, "is_max", "0")):
                self.showMaximized()
                paw.move(0, 0)

    def center_pos(self):
        left = (self.parent().parent().size().width() - self.parent().size().width()) / 2
        top = (self.parent().parent().size().height() - self.parent().size().height()) / 2
        return left, top

    def expand_size(self, paw, sizeBefore, sizeAfter):
        wDelta = 0 if sizeBefore.width() < sizeAfter.width() else sizeBefore.width() - sizeAfter.width()
        hDelta = (
                0 if sizeBefore.height() < sizeAfter.height() else sizeBefore.height() - sizeAfter.height()
            )
        if wDelta or hDelta:
            paw.resize(paw.size().width() + wDelta, paw.size().height() + hDelta)

    def fit_size_and_pos(self, paw):
        """ ensure form fits outside window
        """
        parent_size = paw.parent().size()

        size = paw.size()
        original_size = paw.size()

        pos = paw.pos()
        orginal_pos = paw.pos()
        # width
        if parent_size.width() - (size.width()) < 0:
            size.setWidth(parent_size.width())
        if pos.x() + size.width() > parent_size.width():
            pos.setX(parent_size.width() - size.width())
        if pos.x() < 0:
            pos.setX(0)
        # height
        if parent_size.height() - (size.height()) < 0:
            size.setHeight(parent_size.height())

        if pos.y() + size.height() > parent_size.height():
            pos.setY(parent_size.height() - size.height())
        if pos.y() < 0:
            pos.setY(0)
        if orginal_pos != pos:
            paw.move(pos)
        if size != original_size:
            paw.resize(size)

    def set_position(self, left, top):
        paw = self.parent()
        if paw is not None:
            paw.move(left, top)

    def set_size(self, w, h):
        paw = self.parent()
        if paw is not None:
            paw.resize(w, h)

    def get_position(self):
        parent_mdi_sub_window = self.parent()
        if parent_mdi_sub_window is not None:
            return (parent_mdi_sub_window.pos().x(), parent_mdi_sub_window.pos().y())

    def showEvent(self, event=None):
        if self.shown:
            return

        self.zz_form.form_stack.append(self)

        if self.zz_form.before_form_show() is False:
            self.zz_form.close()
            return

        if not isinstance(self.parent(), QMdiSubWindow):
            self.escapeEnabled = False

        # first_widget = self.widgets[list(self.widgets.keys())[0]]
        # while (
        #     not first_widget.isEnabled()
        #     or (hasattr(first_widget, "isReadOnly") and first_widget.isReadOnly())
        #     or first_widget.focusPolicy() == Qt.NoFocus
        # ):
        #     # print(first_widget)
        #     first_widget = first_widget.nextInFocusChain()

        # first_widget.setFocus()
        # print(first_widget)
        # set focus into first enabled widget
        for widget_name in self.widgets:
            widget = self.widgets[widget_name]
            if widget.focusPolicy() == Qt.NoFocus:
                continue
            if hasattr(widget, "isReadOnly") and widget.isReadOnly():
                continue
            if widget.isEnabled():
                widget.setFocus()
                break

        if self.zz_form.do_not_save_geometry is False:
            self.restore_geometry(zzapp.zz_app.settings)

        self.shown = True
        if event:
            event.accept()
        self.zz_form.after_form_show()

    def keyPressEvent(self, event: QEvent):
        key = event.key()
        keyText = QKeySequence(event.modifiers() | event.key()).toString()
        if key == Qt.Key_Escape and self.escapeEnabled:
            self.close()
        elif key == Qt.Key_Escape and not self.escapeEnabled:
            event.ignore()
            # return
        elif self.mode == "form" and key in (Qt.Key_Up,):
            QApplication.sendEvent(self, QKeyEvent(QEvent.KeyPress, Qt.Key_Tab, Qt.ShiftModifier))
        elif self.mode == "form" and key in (Qt.Key_Enter, Qt.Key_Return, Qt.Key_Down):
            QApplication.sendEvent(self, QKeyEvent(QEvent.KeyPress, Qt.Key_Tab, event.modifiers()))
        elif self.mode == "grid" and key in (Qt.Key_Return,):
            self.zz_form.grid_double_clicked()
        elif keyText in self.hotkey_widgets:  # is it form hotkey
            for widget in self.hotkey_widgets[keyText]:
                if widget.is_enabled() and hasattr(widget, "valid"):
                    widget.valid()
                    return
                    # validate only not hotkeyed widget
                    # if wi != qApp.focusWidget() and \
                    #         hasattr(qApp.focusWidget(), "meta") and \
                    #         qApp.focusWidget().meta.get("key"):
                    #     if qApp.focusWidget().valid() is False:
                    #         return
                    # return wi.valid()

        #     for wi in self.hotKeyWidgets[keyText]:
        #         if wi.isEnabled():
        # else:
        # event.acc5ept()

    def close(self):
        super().close()
        if self.parent() is not None:
            if isinstance(self.parent(), QMdiSubWindow):
                self.parent().close()
        else:
            QDialog.close(self)

    def closeEvent(self, event=None):
        self.zz_form.close()
        if event:
            event.accept()


# Tells the module which engine to use
zzgui.zzdialogs.ZzForm = ZzForm
zzMess
zzWait
zzAskYN



if __name__ == "__main__":
    import sys

    sys.path.insert(0, ".")

    from demo.demo import demo

    demo()

from threading import Thread, current_thread
import time
from zzgui.zzform import ZzForm
import zzgui.zzapp as zzapp


def center_window(form: ZzForm):
    w, h = zzapp.zz_app.main_window.get_size()
    form.form_stack[0].set_size(w * 0.33, h * 0.5)
    form.form_stack[0].set_position(w * 0.33, h * 0.15)


def zzMess(mess="", title="Message"):
    form = ZzForm(title)
    form.do_not_save_geometry = True
    form.add_control("mess", control="text", data=f"{mess}", readonly=True)
    if form.add_control("/h"):
        form.add_control("/s")

        form.add_control(
            "close",
            "Ok",
            control="button",
            eat_enter=True,
            hotkey="PgDown",
            valid=form.close,
        )

        form.add_control("/s")
        form.add_control("/")

    form.before_form_show = lambda: center_window(form)
    form.show_app_modal_form()


def zzAskYN(mess, title="Ask"):
    form = ZzForm(title)
    form.do_not_save_geometry = True
    form.choice = 0
    form.add_control("/")

    form.controls.add_control(
        "mess", control="text", data=f"{mess}", readonly="*", disabled="*"
    )

    if form.add_control("/h"):
        form.add_control("/s")

        def buttonPressed(form=form, answer=0):
            form.choice = answer
            form.close()

        form.add_control(
            "cancel",
            "Cancel",
            control="button",
            valid=lambda: buttonPressed(form, 1),
            eat_enter="*",
        )
        form.add_control(
            "ok",
            "Ok",
            control="button",
            valid=lambda: buttonPressed(form, 2),
            eat_enter="*",
            tag="ok",
            hotkey="PgDown",
        )
        form.add_control("/s")
        form.add_control("/")
    form.before_form_show = lambda: center_window(form)
    form.show_app_modal_form()
    return form.choice


class ZzThread(Thread):
    def __init__(self, group=None, target=None, name=None, args=(), kwargs={}):
        Thread.__init__(self, group, target, name, args, kwargs)
        self._target = target
        self._args = args
        self.min = 0
        self.max = 0
        self.shadow_value = 0
        self.value = 0
        self._return = None
        self.start_time = time.time()

    @staticmethod
    def set_min(value):
        current_thread().min = value

    @staticmethod
    def set_max(value):
        current_thread().max = value

    @staticmethod
    def get_max():
        c_thread = current_thread()
        if hasattr(c_thread, "max"):
            return c_thread.max
        else:
            return 0

    @staticmethod
    def step(step_value=1):
        current_thread().shadow_value += 1
        sv = current_thread().shadow_value
        if sv % step_value == 0:
            current_thread().value = sv

    @staticmethod
    def get_current():
        return current_thread().value

    def time(self):
        return time.time() - self.start_time

    def run(self):
        self._return = self._target(*self._args)


class ZzWaitForm:
    def __init__(self, mess, worker_thread):
        self.tick = {}
        self.worker_thread = worker_thread
        self.wait_window = ZzForm("Wait...")
        self.wait_window.add_control("/s")
        self.wait_window.add_control("/h")
        self.wait_window.add_control("/s")
        self.wait_window.add_control("", label=mess, control="label")
        self.wait_window.add_control("/s")
        self.wait_window.add_control("/")
        steps_count_separator = ""
        if ZzThread.get_max() != 0:
            steps_count_separator = "/"

        if self.wait_window.add_control("/h"):
            self.wait_window.add_control("progressbar", "", control="progressbar")
            self.wait_window.add_control("min", "", control="label")
            self.wait_window.add_control("", steps_count_separator, control="label")
            self.wait_window.add_control("value", "", control="label")
            self.wait_window.add_control("", steps_count_separator, control="label")
            self.wait_window.add_control("max", "", control="label")
            self.wait_window.add_control("time", "", control="label")
        self.wait_window.add_control("/")
        self.wait_window.add_control("/s")
        self.show()

        if self.worker_thread.min != 0:
            self.wait_window.w.progressbar.set_min(self.worker_thread.min)
            self.wait_window.s.min = self.worker_thread.min
        if self.worker_thread.max != 0:
            self.wait_window.w.progressbar.set_max(self.worker_thread.max)
            self.wait_window.s.max = self.worker_thread.max

    def step(self):
        if self.worker_thread.value != 0:
            self.wait_window.w.progressbar.set_value(self.worker_thread.value)
            self.wait_window.s.value = self.worker_thread.value

        thread_time = int(self.worker_thread.time())
        # print(self.worker_thread.time())
        sec = thread_time % 60
        min = (thread_time - sec) % 3600
        hours = thread_time - min * 3600 - sec
        sec = int(sec)
        self.wait_window.s.time = f" Time {hours:02}:{min:02}:{sec:02}"
        zzapp.zz_app.process_events()

    def show(self):
        self.wait_window.show_mdi_form()
        zzapp.zz_app.process_events()
        w, h = zzapp.zz_app.main_window.get_size()
        self.wait_window.form_stack[0].set_size(w * 0.8, h * 0.15)
        self.wait_window.form_stack[0].set_position(w * 0.1, h * 0.3)
        zzapp.zz_app.process_events()

    def close(self):
        self.wait_window.close()
        zzapp.zz_app.process_events()


def zzWaitStep(step_value=1):
    ZzThread.step(step_value)


def zzWaitMax(max_value=0):
    ZzThread.set_max(max_value)


def zzWait(worker, mess=""):
    wait_window = None
    wait_window_on = False
    last_focus_widget = zzapp.zz_app.focus_widget()
    last_progressbar_value = 0
    last_progressbar_time = 0
    zzapp.zz_app.lock()
    worker_thread = ZzThread(target=worker)
    worker_thread.start()
    while worker_thread.is_alive():
        time.sleep(0.3)
        if worker_thread.time() > 1 and wait_window_on is not True:
            wait_window_on = True
            wait_window = ZzWaitForm(mess, worker_thread)
        if wait_window_on is True:
            if worker_thread.min < worker_thread.max:
                if (
                    worker_thread.value != 0
                    and last_progressbar_value != worker_thread.value
                ):
                    wait_window.step()
            elif worker_thread.time() - last_progressbar_time > 1:
                wait_window.step()
                last_progressbar_time = worker_thread.time()
        last_progressbar_value = worker_thread.value
        zzapp.zz_app.process_events()
    zzapp.zz_app.unlock()
    if wait_window is not None:
        wait_window.close()
    if hasattr(last_focus_widget, "set_focus"):
        last_focus_widget.set_focus()

    zzapp.zz_app.show_statusbar_mess(f"{worker_thread.time():.2f}")
    return worker_thread._return



import sys

if __name__ == "__main__":

    sys.path.insert(0, ".")

    from demo.demo import demo

    demo()


from zzgui.zzapp import ZzActions

from PyQt5.QtWidgets import (
    QFrame,
    QHBoxLayout,
    QVBoxLayout,
    QToolBar,
    QLabel,
    QToolButton,
    QSizePolicy,
    QMenu,
)
from PyQt5.QtCore import Qt, QMargins

from zzgui.qt5.zzwidget import ZzWidget
from zzgui.qt5.zzwindow import zz_align


class zztoolbar(QFrame, ZzWidget):
    def __init__(self, meta):
        super().__init__(meta)
        self.setLayout(QVBoxLayout() if "v" in meta.get("control") else QHBoxLayout())
        self.setSizePolicy(QSizePolicy.Policy.Maximum, QSizePolicy.Policy.Maximum)
        self.layout().setAlignment(zz_align["7"])
        self.layout().setSpacing(-1)
        self.layout().setContentsMargins(QMargins(0, 0, 0, 0))
        self.toolBarPanel = QToolBar()

        action_list = []
        if isinstance(meta.get("actions"), ZzActions):
            action_list.extend(meta.get("actions"))
            actions = meta.get("actions")
        elif isinstance(meta.get("actions"), list):
            for x in meta.get("actions"):
                if isinstance(x, ZzActions):
                    action_list.extend(x)
            actions = meta.get("actions")[0]

        if action_list is []:
            return

        tool_bar_qt_actions = QMenu()
        cascade_action = {"": tool_bar_qt_actions}

        for action in action_list:
            if action.get("text", "").startswith("/"):
                continue
            worker = action.get("worker", None)
            action_text_list = action["text"].split("|")
            for x in range(len(action_text_list)):
                action_key = "|".join(action_text_list[:x])
                action_text = action_text_list[x]
                if action_text == "-":
                    action["engineAction"] = cascade_action[action_key].addSeparator()
                else:
                    if x + 1 == len(action_text_list):  # real action

                        action["engineAction"] = cascade_action[action_key].addAction(action_text)
                        action["engineAction"].setToolTip(action.get("mess", ""))
                        action["engineAction"].setStatusTip(action.get("mess", ""))
                        if worker:
                            action["engineAction"].triggered.connect(worker)
                        elif action.get("child_where") and action.get("child_form"):

                            def getChildForm(action):
                                def rd():
                                    self.meta["form"].show_child_form(action)

                                return rd

                            action["engineAction"].triggered.connect(getChildForm(action))

                        action["engineAction"].setShortcut(
                            action["hotkey"] if not action["hotkey"] == "Spacebar" else Qt.Key_Space
                        )
                        action["engineAction"].setShortcutContext(Qt.WidgetWithChildrenShortcut)
                    else:  # cascade
                        subMenu = "|".join(action_text_list[: x + 1])
                        if subMenu not in cascade_action:
                            cascade_action[subMenu] = cascade_action[action_key].addMenu(
                                f"{action_text}  {'' if '|' in subMenu else '  '}"
                            )

        self.main_button = QToolBar()
        self.main_button_action = self.main_button.addAction("☰")
        self.main_button_action.setToolTip(self.meta.get("mess", ""))
        self.main_button_action.setMenu(tool_bar_qt_actions)
        self.main_button.widgetForAction(self.main_button_action).setPopupMode(QToolButton.InstantPopup)
        self.layout().addWidget(self.main_button)
        if actions.show_main_button is False:
            self.main_button.setVisible(False)

        self.toolBarPanel.addSeparator()
        self.toolBarPanel.addActions(tool_bar_qt_actions.actions())
        for x in self.toolBarPanel.actions():
            if hasattr(self.toolBarPanel.widgetForAction(x), "setPopupMode"):
                self.toolBarPanel.widgetForAction(x).setPopupMode(QToolButton.InstantPopup)

        if actions.show_actions:
            self.layout().addWidget(self.toolBarPanel)

    def set_context_menu(self, widget):
        widget.setContextMenuPolicy(Qt.ActionsContextMenu)
        widget.addActions(self.toolBarPanel.actions())



if __name__ == "__main__":

    import sys

    sys.path.insert(0, ".")

    from demo.demo import demo

    demo()

from PyQt5.QtWidgets import QGroupBox, QSplitter, QSizePolicy
from PyQt5.QtCore import Qt


from zzgui.qt5.zzwindow import ZzFrame
from zzgui.qt5.zzwidget import ZzWidget


class zzframe(QGroupBox, ZzWidget, ZzFrame):
    def __init__(self, meta):
        super().__init__(meta)
        ZzFrame.__init__(self, meta.get("name", "/v")[1])
        self.splitter = None
        if meta.get("name", "")[2:3] == "s":  # Splitter!
            self.splitter = zzsplitter()
            if meta.get("name").startswith("/v"):
                self.splitter.setOrientation(Qt.Orientation.Vertical)
            self.layout().addWidget(self.splitter)
        if meta.get("label"):
            self.set_title(meta.get("label"))

    def hide_border(self):
        self.setObjectName("grb")
        self.setStyleSheet("QGroupBox#grb {border:0}")
        self.set_title("")

    def set_title(self, title):
        self.setTitle(title)

    def get_widget_count(self):
        return self.layout().count()

    def add_widget(self, widget=None, label=None):
        if self.splitter is not None:
            self.splitter.addWidget(widget)
            if hasattr(widget, "meta"):
                # print(widget.meta.get("stretch"), widget)
                self.splitter.setStretchFactor(self.splitter.count()-1, widget.meta.get("stretch", 0))
            # if hasattr(widget, "meta"):
            #     if "toolbar" in widget.meta.get("name", ""):
            #         widget.set_context_menu(self.splitter)
        else:
            return super().add_widget(widget=widget, label=label)


class zzsplitter(QSplitter):
    def __init__(self):
        super().__init__()
        self.setSizePolicy(QSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding))

    def get_sizes(self):
        return ",".join([f"{x}" for x in self.sizes()])

    def set_sizes(self, sizes):
        if sizes:
            sizes = [int(x) for x in sizes.split(",")]
            self.setSizes(sizes)



import sys

if __name__ == "__main__":

    sys.path.insert(0, ".")

    from demo.demo import demo

    demo()

from PyQt5.QtWidgets import QLabel


from zzgui.qt5.zzwidget import ZzWidget


class zzlabel(QLabel, ZzWidget):
    def __init__(self, meta):
        super().__init__({"label": meta.get("label", "")})
        self.set_text(self.meta["label"])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zzgui', 'zzgui.qt5', 'zzgui.qt5.widgets']

package_data = \
{'': ['*']}

install_requires = \
['PyQt5>=5.15.6,<6.0.0', 'QScintilla>=2.13.1,<3.0.0']

setup_kwargs = {
    'name': 'zzgui',
    'version': '0.1.12',
    'description': 'Python GUI toolkit',
    'long_description': '# The light Python GUI builder (currently based on PyQt5)\n\n# How to start \n## With docker && x11:\n```bash\ngit clone https://github.com/AndreiPuchko/zzgui.git\n#                      sudo if necessary \ncd zzgui/docker-x11 && ./build_and_run_menu.sh\n```\n## With PyPI package:\n```bash\npoetry new project_01 && cd project_01 && poetry shell\npoetry add zzgui\ncd project_01\npython -m zzgui > example_app.py && python example_app.py\n```\n## Explore sources:\n```bash\ngit clone https://github.com/AndreiPuchko/zzgui.git\ncd zzgui\npip3 install poetry\npoetry shell\npoetry install\npython3 demo/demo_01.py     # basic: main menu, form & widgets\npython3 demo/demo_02.py     # forms and forms in form\npython3 demo/demo_03.py     # grid form (CSV data), automatic creation of forms based on data\npython3 demo/demo_04.py     # progressbar, data loading, sorting and filtering\npython3 demo/demo_05.py     # nonmodal form\npython3 demo/demo_06.py     # code editor\npython3 demo/demo_07.py     # database app (4 tables, mock data loading) - requires a zzdb package\npython3 demo/demo_08.py     # database app, requires a zzdb package, autoschema\n```\n\n## demo/demo_03.py screenshot\n![Alt text](https://andreipuchko.github.io/zzgui/screenshot.png)\n# Build standalone executable \n(The resulting executable file will appear in the folder  dist/)\n## One file\n```bash\npyinstaller -F demo/demo.py\n```\n\n## One directory\n```bash\npyinstaller -D demo/demo.py\n```\n',
    'author': 'Andrei Puchko',
    'author_email': 'andrei.puchko@gmx.de',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7',
}


setup(**setup_kwargs)



from zzgui.qt5.widgets import (
    zzlabel,
    zzline,
    zzframe,
    zzcheck,
    zztoolbar,
    zztab,
    zztext,
    zzcode,
    zzbutton,
    zzrelation,
    zzradio,
    zzlookup,
    zzcombo,
    zztoolbutton,
    zzprogressbar,
    zzgrid,
    zzdate,
    zzspace,
    zzlist,
    zzspin,
)


(
    zzlabel,
    zzline,
    zzframe,
    zzcheck,
    zztoolbar,
    zztab,
    zztext,
    zzcode,
    zzbutton,
    zzrelation,
    zzradio,
    zzlookup,
    zzcombo,
    zztoolbutton,
    zzprogressbar,
    zzgrid,
    zzdate,
    zzspace,
    zzlist,
    zzspin,
)



if __name__ == "__main__":
    import sys

    sys.path.insert(0, ".")

    from demo.demo import demo

    demo()


from PyQt5.QtWidgets import QWidget
from PyQt5.QtGui import QFontMetrics

from zzgui import zzwidget
from zzgui.qt5.zzwindow import zz_align


class ZzWidget(QWidget, zzwidget.ZzWidget):
    def __init__(self, meta):
        super().__init__()
        zzwidget.ZzWidget.__init__(self, meta)
        # self.setContentsMargins(0, 0, 0, 0)

    def set_tooltip(self, mess):
        self.setToolTip(mess)

    def set_disabled(self, arg=True):
        self.setEnabled(True if not arg else False)

    def set_enabled(self, arg=True):
        self.setEnabled(True if arg else False)

    def set_text(self, text):
        if hasattr(self, "setText"):
            self.setText(f"{text}")

    def get_text(self):
        if hasattr(self, "text"):
            return self.text()
        return ""

    def set_readonly(self, arg):
        if hasattr(self, "setReadOnly"):
            self.setReadOnly(True if arg else False)

    def is_enabled(self):
        return self.isEnabled()

    def set_visible(self, arg=True):
        self.setVisible(arg)

    def is_readonly(self):
        if hasattr(self, "isReadOnly"):
            return self.isReadOnly()

    def set_focus(self):
        self.setFocus()

    def set_maximum_width(self, width):
        if self.meta.get("control", "") not in ("radio", "check"):
            self.setMaximumWidth(QFontMetrics(self.font()).width("W") * width)

    def set_maximum_len(self, length):
        if hasattr(self, "setMaxLength"):
            return self.setMaxLength(length)

    def set_alignment(self, alignment):
        if hasattr(self, "setAlignment"):
            self.setAlignment(zz_align[f"{alignment}"])

    def valid(self):
        if self.meta.get("valid"):
            return self.meta.get("valid", lambda: True)()
        else:
            return True

    def when(self):
        if self.meta.get("when"):
            return self.meta.get("when", lambda: True)()
        else:
            return True
